#Material Maintenance Sketchup Plugin
#Copyright (C) <2012>  <www.KeepingMyHandIn.com>
#  
#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.
#
#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.
#
#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

require 'sketchup.rb'
require 'extensions.rb'
require 'LangHandler.rb'

module MH_KeepingMyHandIn

  unless file_loaded?( __FILE__ )

    UI.menu("Plugins").add_item("Material Maintenance"){ self.materialMaintenance };
    
  end

  class MaterialMaintenance

    # Materials to process
    @materialsMap = nil;

    # Find Variables
    @layerMap = nil;
    @constructionList = nil;
    @constructionColor = nil;
    
    # Dialog
    @my_dialog = nil;
    
    # Temp folder
    @tempDir = nil;
    
    @eCount = nil;
    
    @jSon = nil;

    
    
  ##################################################################################################  
  #                                            UI Handlers                                         #
  ##################################################################################################

    def initialise()

      @my_dialog = UI::WebDialog.new("Material Maintenance", false, nil, 500, 200, 100, 100, true);
      
      @my_dialog.min_width = 550;
      @my_dialog.min_height = 800;

      # Find and show our html file
      html_path = File.expand_path(File.dirname(__FILE__)) + "/index.html";
      @my_dialog.set_file(html_path);
      @my_dialog.show();

      @my_dialog.add_action_callback("fromUI") do |web_dialog, parameterString|
        fromUIHandler(web_dialog, parameterString);
      end
      
      # Find and clear the temp directory.
      tempPath = File.expand_path( ENV['TMPDIR'] || ENV['TMP'] || ENV['TEMP'] );
      @tempDir = File.join( tempPath, 'material_maintenance' );
      
      if File.exist?( @tempDir )
          
        File.delete(*Dir.glob(s = @tempDir + "/*.png"));
     
      else
      
        Dir.mkdir( @tempDir );
          
      end
      
      @jSon = JSON.new;
    
    end
    
    def getSystemParameters()
      
      puts "!!!!! - IN getSystemParameters !!!!!";
    
      return [@tempDir];
    end
    
    def fromUIHandler(web_dialog, parameterString)
    
      puts "!!!=> fromUIHandler: parameter string = #{parameterString}";
      
      params = unmarshalParams(parameterString);
      
      params.each { |p|
        p.puts "p1 = >>#{p}<<";
      }
      
      metaParams = Array.new();
      metaParams[0] = params.shift();
      metaParams[1] = params.shift();
      targetFunction = params.shift();
            
      n = params.length;
      
      for i in 0..n-1
        if params[i][0].chr == 183.chr
            params[i] = unmarshalArray(params[i]);
        end
      end
      
      puts "calling function >>#{targetFunction}<<";
      
      results = self.send(targetFunction.to_sym, *params);
     
      if (results == nil || results == true)
        results = [];
      end

      n = results.length;
      for i in 0..n-1
        if results[i].kind_of?(Array)
            results[i] = marshalArray(results[i]);
        end
      end
      
      results = marshalParams(metaParams.concat(results));
    
      js_command = "material_maintenance.Sketchup.callUI('"+ results + "')";
      web_dialog.execute_script(js_command);
        
    end
    
    def marshalArray(a)
        return 183.chr + a.join(183.chr);
    end
    
    def unmarshalArray(s)
        temp = s.split(183.chr);
        temp.shift();
        return temp;
    end
    
    def marshalParams(p)
        return p.join(182.chr);
    end
    
    def unmarshalParams(s)
          return s.split(182.chr);
    end
    
    def updateStatus(p)
      
    end;
    
  ################################### END ###########################################  
    
    def refreshMaterials(selectionOnly, wholeModel, components)
    
      puts "!!!!! - IN refreshMaterials !!!!!";

      @materialsMap = Hash.new;
      
      model = Sketchup.active_model;
      
      @eCount = 0;
      
      start_time = Time.now
  
      if (selectionOnly == "true")
      
        if (!model.selection.empty?)
          
          model.selection.each do |e| 
          
            @eCount += 1;
          
            navigateEntityGraph(e, nil);
            
          end
         
        end
      
      elsif (wholeModel == "true")
      
          model.entities.each do |e|

            @eCount += 1;

            navigateEntityGraph(e, nil);
            
          end
      
      end 
      
      if (components == "true")
      
        model.definitions.each do |d|
        
          next if d.image?;
          
          d.entities.each do |e|
          
            @eCount += 1;

            navigateEntityGraph(e, d);
            
          end
          
        end  
      
      end
      
      end_time = Time.now

      puts "String Time = #{@time_taken = end_time - start_time}"
      
      puts "Total Entities = #{@eCount}";
       
      File.delete(*Dir.glob(s = @tempDir + "/*.png"));
      
      Sketchup.active_model.materials.each do |m|
      
        m.write_thumbnail(@tempDir + "/" + m.display_name.gsub(Regexp.new('[/\\?%*:|"<>.]'), 182.chr) + "_" + m.object_id.to_s + '.png', 32);
        
      end
      
      matOldIdsAndNames = getMaterialIdsAndNames(@materialsMap.keys);
      matAllIdsAndNames = getMaterialIdsAndNames(Sketchup.active_model.materials.collect {|m| m});

      return [matOldIdsAndNames[0], matOldIdsAndNames[1], matAllIdsAndNames[0], matAllIdsAndNames[1]];
    
    end
    
    def getMaterialIdsAndNames(materials)
    
      materials.delete(:"nil");
    
      sortedList = materials.sort{|x,y| x.display_name.upcase <=> y.display_name. upcase};

      return [sortedList.collect {|m| m.object_id.to_s}, sortedList.collect {|m| m.display_name}]
      
    end
    
    def convertMaterialIdsToObjects(selectedMaterials)
      
      return selectedMaterials.collect {|id| ObjectSpace._id2ref(Integer(id)) };
    
    end
    
    def findMaterialWithConstructionLines (selectedMaterials)
      
      materials = @materialsMap.values_at(*convertMaterialIdsToObjects(selectedMaterials));
      
      activeEntities = Sketchup.active_model.active_entities;
      
      allEntities = [];
      
      materials.each do |entitiesTypesHash|
      
        if (entitiesTypesHash != nil)
     
          entitiesTypesHash.each do |typeName, entitiesArray|
          
            if (entitiesArray != nil)
            
              allEntities.concat(entitiesArray);
            
            end  
          
          end
          
        end
        
      end
      
      allEntities.uniq!();

      Sketchup.active_model.start_operation("Find with construction lines", true);
      
      # Construction Lines
      @constructionList = Array.new;
      @constructionColor = Sketchup.active_model.rendering_options["ConstructionColor"];
      Sketchup.active_model.rendering_options["ConstructionColor"] = "red";
      
      allEntities.each do |e|
        
        if ((e.is_a? Sketchup::ComponentInstance) || (e.is_a? Sketchup::Group))

          @constructionList.push( activeEntities.add_cline( e.bounds.center, Geom::Vector3d.new(0, 0, 1)) );
          @constructionList.push( activeEntities.add_cline( e.bounds.center, Geom::Vector3d.new(0, 1, 0)) );
          @constructionList.push( activeEntities.add_cline( e.bounds.center, Geom::Vector3d.new(1, 0, 0)) );
          @constructionList.push( activeEntities.add_cpoint( e.bounds.center) );
          
        elsif (e.is_a? Sketchup::Face)
        
          @constructionList.push(e.parent.entities.add_cline( e.bounds.center.offset(e.normal, -20), e.bounds.center.offset(e.normal, 20) ) );
          @constructionList.push(e.parent.entities.add_cpoint( e.bounds.center));
          
        else
        
          @constructionList.push(e.parent.entities.add_cpoint( e.bounds.center));
          
        end
        
      end
      
      Sketchup.active_model.commit_operation;
    
    end
    
    def clearMaterialWithConstructionLines()
    
      activeEntities = Sketchup.active_model.active_entities;
    
      Sketchup.active_model.start_operation("Clear with construction lines", true);

      activeEntities.erase_entities(@constructionList);
      
      Sketchup.active_model.rendering_options["ConstructionColor"] = @constructionColor;

      Sketchup.active_model.commit_operation;
      
      @constructionList.clear();
      @constructionList = nil;
    
    end
    
    def findMaterialWithLayers (selectedMaterials)
      
      materials = @materialsMap.values_at(*convertMaterialIdsToObjects(selectedMaterials));

      allEntities = [];
      
      materials.each do |entitiesTypesHash|
      
        if (entitiesTypesHash != nil)
     
          entitiesTypesHash.each do |typeName, entitiesArray|
          
            if (typeName == :"Face" || typeName == :"Face_Back" || typeName == :"ComponentInstance" || typeName == :"Group")
          
              if (entitiesArray != nil)
              
                allEntities.concat(entitiesArray);
              
              end  
          
            end
          
          end
          
        end
        
      end
      
      allEntities.uniq!();
      
      Sketchup.active_model.start_operation("Find with layers", true);
      
      # Use layer 
      @layerMap = Hash.new;
      findLayer = Sketchup.active_model.layers["__FINDER__"];
      if (findLayer == nil)
        findLayer = Sketchup.active_model.layers.add("__FINDER__");
      end
      findLayer.visible = true;
      Sketchup.active_model.rendering_options["DisplayColorByLayer"] = true;

      allEntities.each do |e|
        
          @layerMap[e] = e.layer;
          
          e.layer = findLayer;
        
      end
      
      Sketchup.active_model.commit_operation;
      
    end
    
    def clearMaterialWithLayers()
    
      Sketchup.active_model.start_operation("Clear with layers", true);
      
      @layerMap.each do |e, l|
      
        e.layer = l;
        
      end

      Sketchup.active_model.rendering_options["DisplayColorByLayer"] = false;

      Sketchup.active_model.commit_operation;
      
      @layerMap.clear();
      @layerMap = nil;
    
    end

    def selectMaterialInModel (selectedMaterials)
      
      materials = @materialsMap.values_at(*convertMaterialIdsToObjects(selectedMaterials));
      
      allEntities = [];
      
      materials.each do |entitiesTypesHash|
      
        if (entitiesTypesHash != nil)
     
          entitiesTypesHash.each do |typeName, entitiesArray|
          
            if (entitiesArray != nil)
            
              allEntities.concat(entitiesArray);
            
            end  
          
          end
          
        end
        
      end
      
      allEntities.uniq!();
      
      selection = Sketchup.active_model.selection;

      allEntities.each do |e|
      
        selection.add(e);
        
      end
      
    end
    
    def clearSelectedMaterials()

      Sketchup.active_model.selection.clear();
    
    end

    def findComponents(selectedMaterials)
    
      materials = @materialsMap.values_at(*convertMaterialIdsToObjects(selectedMaterials));
      
      components = [];
      
      materials.each do |entitiesTypesHash|
      
        if (entitiesTypesHash != nil)
        
          if (entitiesTypesHash[:"Component"] != nil)
            
              components.concat(entitiesTypesHash[:"Component"]);
            
          end
          
        end
        
      end
      
      File.delete(*Dir.glob(s = @tempDir + "/*.comp.png"));

      components.each do |c|

        c.save_thumbnail(@tempDir + "/" + c.name.gsub(Regexp.new('[/\\?%*:|"<>.]'), 182.chr) + "_" + c.object_id.to_s + '.comp.png');
        
      end
      
      return [@jSon.marshal(components.collect{|c| [c.object_id.to_s, c.name, c.count_instances]})];
    
    end
    
    def refreshComponentDefinitions()
    
      File.delete(*Dir.glob(s = @tempDir + "/*.comp.png"));

      Sketchup.active_model.definitions.each { |c|

        c.save_thumbnail(@tempDir + "/" + c.name.gsub(Regexp.new('[/\\?%*:|"<>.]'), 182.chr) + "_" + c.object_id.to_s + '.comp.png');
        
      }
    
      return [@jSon.marshal(Sketchup.active_model.definitions.collect{|c| [c.object_id.to_s, c.name, c.count_instances]})];
    
    end
    
    def findComponentParentsFromId(compId)
      findComponentParents(ObjectSpace._id2ref(Integer(compId)));
    end
    
    def findComponentParents(comp)
    
#      puts "==> level 1 #{comp.name}";

      comp.instances.each { |inst|
      
        parent = inst.parent;
      
        while !((parent.is_a? Sketchup::DefinitionList) || (parent.is_a? Sketchup::Model));
          
          if (parent.is_a? Sketchup::ComponentDefinition)
          
            puts parent.name;
            
            findComponentParents(parent);
            
          end
          
          parent = parent.parent();

        end
        
        if (parent.is_a? Sketchup::Model)
          puts "Model"
        end
      
      }
      
 #     puts "<== level 1 #{comp.name}";
      
    end
    
    def convertMaterials( oldMaterialType, oldMaterialList, components, newMaterialType, newMaterialId, typeNames)
    
      start_time = Time.now
    
      if oldMaterialType == "Default"

        oldMaterials = [@materialsMap[:"nil"]];
        
      elsif oldMaterialType == "List"

        oldMaterials = @materialsMap.values_at(*convertMaterialIdsToObjects(oldMaterialList));
      
      else
      
        oldMaterials = @materialsMap.values;
          
      end
     
      if newMaterialType == "Default"
      
        newMaterial = nil;
      
      else

        newMaterial = ObjectSpace._id2ref(Integer(newMaterialId));
      
      end
      
      # Convert type strings to symbols
      types = typeNames.collect {|name| name.to_sym}; 
          
      Sketchup.active_model.start_operation("Convert Materials", true);

      replaceMaterials( oldMaterials, newMaterial, types);

      # Update the component thumbs
      oldMaterials.each do |entitiesTypesHash|
           
        compList = entitiesTypesHash[:"Component"];
        
        if ( compList != nil)
        
          compList.each do |comp|
            
            comp.refresh_thumbnail();  # if comp.respond_to?( :refresh_thumbnail );
          
          end
          
        end
        
      end

      Sketchup.active_model.commit_operation;
      
      updateStatus("Done");
      
      end_time = Time.now

      puts "String Time = #{@time_taken = end_time - start_time}"
      
      return [];
      
    end
    
    def replaceMaterials( oldMaterials, newMaterial, changeTypes)

      oldMaterials.each do |typeHash|
      
        changeTypes.each do |type|
        
          if (typeHash[type] != nil)
          
            typeHash[type].each do |e|
            
              if (type == :"Face_Back")
             
                e.back_material = newMaterial;
                
              else

                e.material = newMaterial;
              
              end
              
            end
          
          end  
        
        end
        
      end
    
    end
    
    def processEntity(entity, component)
    
          if (entity.is_a? Sketchup::Drawingelement)
      
        if (entity.material != nil)
      
          if (@materialsMap[entity.material] == nil)
          
            @materialsMap[entity.material] = Hash.new;
            
          end
          
          if (@materialsMap[entity.material][entity.typename.to_sym] == nil)
          
            @materialsMap[entity.material][entity.typename.to_sym] = Array.new;
            
          end
        
          # add the entity
          @materialsMap[entity.material][entity.typename.to_sym].push(entity);
          
          if (component != nil)
            if (@materialsMap[entity.material][:"Component"] == nil)
              @materialsMap[entity.material][:"Component"] = Array.new;
            end
            
            unless (@materialsMap[entity.material][:"Component"].include?(component))
              @materialsMap[entity.material][:"Component"].push(component);
            end          
          
          end
          
        else
          
          if (@materialsMap[:"nil"] == nil)
          
            @materialsMap[:"nil"] = Hash.new;
            
          end
          
          if (@materialsMap[:"nil"][entity.typename.to_sym] == nil)
          
            @materialsMap[:"nil"][entity.typename.to_sym] = Array.new;
            
          end

          # add the entity
          @materialsMap[:"nil"][entity.typename.to_sym].push(entity);

        end
        
      end  

      if (entity.is_a? Sketchup::Face)
        
        if (entity.back_material != nil)
      
          if (@materialsMap[entity.back_material] == nil)
          
            @materialsMap[entity.back_material] = Hash.new;
            
          end
        
          if (@materialsMap[entity.back_material][:"Face_Back"] == nil)
          
            @materialsMap[entity.back_material][:"Face_Back"] = Array.new;
            
          end

          # add the entity
          @materialsMap[entity.back_material][:"Face_Back"].push(entity);
          
          if (component != nil)
            if (@materialsMap[entity.back_material][:"Component"] == nil)
              @materialsMap[entity.back_material][:"Component"] = Array.new;
            end
            
            unless (@materialsMap[entity.back_material][:"Component"].include?(component))
              @materialsMap[entity.back_material][:"Component"].push(component);
            end          
          
          end
        
        else
      
          if (@materialsMap[:"nil"] == nil)
          
            @materialsMap[:"nil"] = Hash.new;
            
          end
        
          if (@materialsMap[:"nil"][:"Face_Back"] == nil)
          
            @materialsMap[:"nil"][:"Face_Back"] = Array.new;
            
          end
          
          # add the entity
          @materialsMap[:"nil"][:"Face_Back"].push(entity);

        end
        
      end

    end
    
    def navigateEntityGraph(entity, component)
    
      stack = Array.new;
      
      stack.push(entity);
      
      while (!stack.empty?)
      
        e = stack.pop();
        
        processEntity(e, component);
      
        if (e.is_a? Sketchup::Group)
        
          e.entities.each do |child|

            @eCount += 1;
          
            stack.push(child);

          end

        end
      
      end

    end
    
    def recurseEntity(entity, component)
    
      processEntity(entity, component);

      if (entity.is_a? Sketchup::Group)
        
        entity.entities.each do |e|
        
          @eCount += 1;
        
          recurseEntity(e, component);
          
        end

      end

    end
    
  end
  
  class JSON

    def marshal( v )
      
      json = "";
      
      marshalValue(json, v);
      
    end

    def marshalValue(json, v)
    
      if (v.is_a? String)
      
        marshalString(json, v);
      
      elsif (v.is_a? Array)
      
        marshalArray(json, v);
      
      elsif (v.is_a? Hash)
      
        marshalArray(json, v);
      
      elsif (v.is_a? Numeric)
        
        json << v.to_s;
      
      elsif (v.is_a? TrueClass)
      
        json << "true";
      
      elsif (v.is_a? FalseClass)
      
        json << "false";
      
      else 
      
        if (v == nil)
        
          json << "null";
          
        else
        
          marshalObject(json, v);
          
        end
        
      end
      
      return json;
        
    end
    
    def marshalString(json, v)
    
      json << '"' << escape(v) << '"';
          
      return json;
    
    end

    def marshalObject(json, a)
    
      json << '{';
      
      vars = a.instance_variables;
       
      vars.each { |m|
      
        marshalString(json, m);
      
        json << ':';
        
        marshalValue(json, a.instance_variable_get(m));

        json << ',';
        
      
      }

      if (vars.size() > 0)
      
        json.chop!();
        
      end
      
      json << '}';

      return json;
    
    end

    def marshalArray(json, a)
    
      json << '[';

      a.each { |e|
      
        marshalValue(json, e);
        
        json << ',';
        
      }
      
      if (a.size() > 0)
      
        json.chop!();
        
      end
      
      json << ']';
      
      return json;
    
    end

    def escape(value)

      ret = "";
      
      value.split("").each do |c|
        
        if (/["\\\/\b\f\n\r\t]/.match( c ) )
        
          ret << '\\' << c;
          
        else
        
          ret << c;
          
        end
        
      end
      
      return ret;
      
    end
    
  end
    
  def self.materialMaintenance

    MaterialMaintenance.new.initialise();
    
  end

  file_loaded( __FILE__ );

end